#!/usr/bin/env python

# -----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as
# authorised by a licensing agreement from ARM Limited
# (C) COPYRIGHT 2018-2019 ARM Limited
# ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorised
# copies and copies may only be made to the extent permitted
# by a licensing agreement from ARM Limited.
# -----------------------------------------------------------------------------

import sys, os, subprocess, logging
import re

invalid_triple = '-1.-1.-1'

# list of versions supported for tools
valid_tool_versions = {
		'gcc_host': { 'min': '4.9.0', 'max': invalid_triple},
		'gcc_target': { 'min': '4.9.0', 'max': invalid_triple},
		'clang': { 'min': '3.8.0', 'max': invalid_triple},
		'go': { 'min': '1.10.0', 'max': invalid_triple},
		'python': { 'min': '2.7.12', 'max': '3.7.1'},
		'ply' : { 'min': '3.6.0', 'max': invalid_triple},
		'ninja' : { 'min': '1.8.0', 'max': invalid_triple},
	}

# given a dot separated triple, return the three integers
def get_triple( rTriple ):
	lTokens = rTriple.split('.')
	iMajor = int(lTokens[0])
	iMinor = int(lTokens[1])
	# Cover all combinations of M.ma, some M.ma.x, and some M.ma.mb
	# If we can't decode the 3rd digit safely - treat it as 0
	try:
		iMicro = int(lTokens[2])
		return (iMajor, iMinor, iMicro)
	except:
		return (iMajor, iMinor, 0)

#compare version triple with range from min to max
def compare_triples( lVer, lMin, lMax ):
	# check for the minimum version required
	if (lVer < lMin):
		return False

	# check for the maximum version required
	if(lMax != get_triple(invalid_triple) and lVer > lMax):
		return False

	return True

# get gcc version triple
def get_gcc_version( rCC ):
	# --dumpversion returns only major number in newer versions of gcc; -dumpfullversion works like -dumpversion in
	# older versions of gcc, but those older versions don't support -dumpfullversion...
	sVersion = invalid_triple
	with open(os.devnull, 'w') as devnull:
		try:
			sVersion = subprocess.check_output([rCC, '-dumpfullversion'], stderr=devnull)
			sVersion = sVersion.decode(sys.getdefaultencoding())
		except subprocess.CalledProcessError:
			try:
				sVersion = subprocess.check_output([rCC, '-dumpversion'], stderr=devnull)
				sVersion = sVersion.decode(sys.getdefaultencoding())
			except subprocess.CalledProcessError:
				#do nothing as already assigned to invalid triple
				pass
		except:
			#do nothing as already assigned to invalid triple
			pass
	return sVersion.rstrip('\n\r ')

# get clang version triple
def get_clang_version( rCLANG ):
	sVersion = invalid_triple
	with open(os.devnull, 'w') as devnull:
		try:
			stdout = subprocess.check_output([rCLANG, '--version'], stderr=devnull)
			stdout = stdout.decode(sys.getdefaultencoding())
			m = re.search(r"version (\d+\.\d+(\.\d+)?)", stdout)
			if m:
				sVersion = m.group(1)
			else:
				logging.warning("Unhandled %s version: \"%s\"" % (rCLANG, stdout.strip()))
		except:
			#do nothing as already assigned to invalid triple
			pass
	return sVersion

# get go version triple
def get_go_version():
	sVersion = invalid_triple
	with open(os.devnull, 'w') as devnull:
		try:
			stdout = subprocess.check_output(['go', 'version'], stderr=devnull)
			stdout = stdout.decode(sys.getdefaultencoding())
			m = re.search(r"go(\d+\.\d+(\.\d+)?) ", stdout)
			if m:
				sVersion = m.group(1)
			else:
				logging.warning("Unhandled go version: \"%s\"" % stdout.strip())
		except:
			#do nothing as already assigned to invalid triple
			pass
	return sVersion

# get ninja version triple
def get_ninja_version():
	sVersion = invalid_triple
	with open(os.devnull, 'w') as devnull:
		try:
			stdout = subprocess.check_output(['ninja', '--version'], stderr=devnull)
			stdout = stdout.decode(sys.getdefaultencoding())
			sVersion = stdout.rstrip('\n\r ')
		except:
			#do nothing as already assigned to invalid triple
			pass
	return sVersion

# get python version triple
def get_python_version():
	imaj = sys.version_info[0]
	imin = sys.version_info[1]
	imicro = sys.version_info[2]
	return str(imaj) + "." + str(imin) + "." + str(imicro)

# get python-ply version triple
def get_ply_version():
	try:
		import ply
		return ply.__version__
	except ImportError:
		return invalid_triple
	except AttributeError:
		logging.warning("python-ply does not have __version__ attribute to check")
		return invalid_triple

# check individual tool version against min and max
def check_version( sToolName, sVer, versionMinMax ):
	lVersion = get_triple(sVer)
	lVerMin = get_triple(versionMinMax["min"])
	lVerMax = get_triple(versionMinMax["max"])
	if lVersion[0] == -1:
		logging.warning( sToolName + " cannot be found" )
	elif compare_triples( lVersion, lVerMin, lVerMax ) != True:
		warning = "Incorrect " + sToolName + " version " + sVer + ", version must be "
		if lVerMax[0] != -1:
			warning += "in the range " + versionMinMax["min"] + " to " + versionMinMax["max"] + "."
		else:
			warning += "higher than " + versionMinMax["min"] + "."
		logging.warning( warning )
		return 1
	else:
		logging.info( sToolName + " version " + sVer + " OK." )
	return 0

# run through all tools
def check_tools_versions( compiler ):

	iResult = 0

	# clang compiler - same used for host and target
	if compiler == "":
		#assume host explore is false so don't check compiler
		logging.info( "No compiler selected" )
	elif "clang" in compiler:
		sVer = get_clang_version(compiler)
		iResult += check_version( compiler, sVer, valid_tool_versions["clang"])
	else: # check target and host for gcc
		sVer = get_gcc_version(compiler)
		iResult += check_version( compiler, sVer, valid_tool_versions["gcc_target"])
		sVer = get_gcc_version("gcc")
		iResult += check_version( "gcc", sVer, valid_tool_versions["gcc_host"])

	#python
	sVer = get_python_version()
	iResult += check_version( "python", sVer, valid_tool_versions["python"])

	#python-ply
	sVer = get_ply_version()
	iResult += check_version( "python-ply", sVer, valid_tool_versions["ply"])

	#ninja
	sVer = get_ninja_version()
	iResult += check_version( "ninja", sVer, valid_tool_versions["ninja"])

	#go
	sVer = get_go_version()
	iResult += check_version( "go", sVer, valid_tool_versions["go"])

	if iResult != 0:
		logging.warning( "One or more of the tools is incompatible with the build system." )
		return False

	return True


def main():
	import argparse

	logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)

	parser = argparse.ArgumentParser()
	parser.add_argument("compiler", type=str, default="", nargs="?", help="The target compiler.")
	args = parser.parse_args()

	if check_tools_versions( args.compiler ) == False:
		return 1
	else:
		return 0

if __name__ == "__main__":
	main()
